var inherits = require('inherits');
var Writable = require('readable-stream').Writable;
var EventEmitter = require('events').EventEmitter;
var copy = require('shallow-copy');

var parseTag = require('./lib/parse_tag.js');
var parseSelector = require('./lib/selector.js');
var match = require('./lib/match.js');
var Tag = require('./lib/tag.js');

var special = (function () {
    var tags = [
        'area', 'base', 'basefont', 'br', 'col', 'hr', 'input',
        'img', 'link', 'meta'
    ];
    var special = {};
    for (var i = 0; i < tags.length; i++) {
        special[tags[i]] = true;
    }
    return special;
})();

inherits(Selector, Writable);
module.exports = Selector;

function Selector (sel, cb) {
    if (!(this instanceof Selector)) return new Selector(sel, cb);
    Writable.call(this, { objectMode: true });
    this.cb = cb;
    this.selector = parseSelector(sel);
    this.stack = [];
    this.matches = [];
    this.streams = [];
}

Selector.prototype._write = function (row, enc, next) {
    var type = row[0], buf = row[1];
    if (typeof buf === 'string') buf = Buffer(buf);
    if (type === 'open') {
        var tag = parseTag(buf);
        this._push(tag);
        
        if (special.hasOwnProperty(tag.name)) {
            this._pop();
        }
    }
    for (var i = 0; i < this.streams.length; i++) {
        var s = this.streams[i];
        var closing = type === 'close'
            && s._row.index === this.stack.length - 1
        ;
        s._writes ++;
        if ((s._writes > 1 || s._outer) && (s._outer || !closing)) {
            s.push(buf);
        }
    }
    if (type === 'close') {
        var tag = parseTag(buf);
        this._pop();
    }
    next();
};

Selector.prototype._push = function (tag) {
    var self = this;
    var row = {
        tag: tag,
        matches: [],
        streams: [],
        index: this.stack.length
    };
    
    for (var i = 0; i < this.matches.length; i++) {
        var m = this.matches[i];
        if (m.stopIndex && m.stopIndex < this.stack.length) continue;
        m.stopIndex = undefined;
        
        var sel = this.selector[m.index];
        if (sel === '>') {
            sel = this.selector[++m.index];
        }
        var prev = m.index > 0 && this.selector[m.index-1];
        
        if (match(sel, tag)) {
            if (++ m.index === this.selector.length) {
                m.index --;
                var t = this._fromTag(tag);
                t.on('stream', function (s) {
                    s._row = row;
                    row.streams.push(s);
                    self.streams.push(s);
                });
                
                this.cb(t);
                break;
            }
        }
        else if (prev === '>') {
            m.stopIndex = this.stack.length;
        }
    }
    
    if (match(this.selector[0], tag)) {
        if (this.selector.length === 1) {
            var t = this._fromTag(tag);
            t.on('stream', function (s) {
                s._row = row;
                row.streams.push(s);
                self.streams.push(s);
            });
            this.cb(t);
        }
        else {
            var m = {
                index: 1,
                startIndex: this.stack.length
            };
            this.matches.push(m);
            row.matches.push(m);
        }
    }
    this.stack.push(row);
};

Selector.prototype._pop = function () {
    var s = this.stack.pop();
    for (var i = 0; i < s.matches.length; i++) {
        var ix = this.matches.indexOf(s.matches[i]);
        if (ix >= 0) this.matches.splice(ix, 1);
    }
    for (var i = 0; i < s.streams.length; i++) {
        s.streams[i].push(null);
        var ix = this.streams.indexOf(s.streams[i]);
        if (ix >= 0) this.streams.splice(ix, 1);
    }
};

Selector.prototype._fromTag = function (tag) {
    return new Tag(tag);
};
